//===- rewriter_ex.cpp ----------------------------------------------------===//
//
//                     The LLVM Compiler Infrastructure
//
// Owner: Bogdan-Andrei Tabacaru
// E-Mail: bogdan.tabacaru88@yahoo.com
//
//===----------------------------------------------------------------------===//
//
// Compile AST changes in place without generating a new file.
//
//===----------------------------------------------------------------------===//

#include "clang/Frontend/FrontendPluginRegistry.h"
#include "clang/AST/AST.h"
#include "clang/AST/RecursiveASTVisitor.h"
#include "clang/AST/ASTConsumer.h"
#include "clang/Frontend/CompilerInstance.h"
#include "llvm/Support/raw_ostream.h"
#include "clang/Rewrite/Core/Rewriter.h"

#include "clang/Frontend/MultiplexConsumer.h"

using namespace clang;

namespace {

Rewriter *rewriter = new Rewriter;

class rewriter_exVisitor : public RecursiveASTVisitor<rewriter_exVisitor> {
public:
  bool VisitFieldDecl(FieldDecl *Declaration) {
    RecordDecl *RD = Declaration->getParent();
    CXXRecordDecl *CxxRD = dyn_cast<CXXRecordDecl>(RD);
    // Only Dump struct A's info
    if (CxxRD && CxxRD->getNameAsString() == "A") {
      Declaration->dump();
      
      QualType QT = Declaration->getTypeSourceInfo()->getType();
      // Only affect the object "value" of type Original
      if (QT.getAsString() == "struct Original") {
        std::string value = "New ";
	// Change field from "Original value" to "New value"
        std::string newValue = value + Declaration->getNameAsString();
      
        llvm::outs() << "\t type name: " << QT.getAsString() << "\n"
                     << "\t variable name: " << Declaration->getNameAsString() << "\n"
                     << "\t C++ class name: " << CxxRD->getNameAsString() << "\n";
        bool done = rewriter->ReplaceText(Declaration->getSourceRange(), newValue); 
	// Buffer prints changes info successfully    
        rewriter->getEditBuffer(rewriter->getSourceMgr().getMainFileID()).write(llvm::outs());
        llvm::outs() << "Rewritten " << (done ? "un":"") << "successfully to " << newValue << "\n";
      }
    }
    return true;
  }
};

class rewriter_exConsumer : public clang::ASTConsumer {
public:
  // Override the constructor in order to pass CI
  explicit rewriter_exConsumer(CompilerInstance *CI)
  : Visitor(new rewriter_exVisitor()), CI(CI)
  { }
  
  // Put this plugin before CodeGen
  void Initialize(ASTContext& Context) {
    rewriter->setSourceMgr(Context.getSourceManager(), Context.getLangOpts());
    // We need to reorder the consumers in the MultiplexConsumer.
    MultiplexConsumer& multiplex = static_cast<MultiplexConsumer&>(CI->getASTConsumer());
    std::vector<ASTConsumer*>& consumers = multiplex.getConsumers();
    ASTConsumer* lastConsumer = consumers.back();
    bool decide = lastConsumer == this;
    llvm::outs() << "this ASTConsumer " << (decide ? "=" : "!=") << " lastConsumer\n";
    ASTConsumer* CodeGen = consumers.front();
    CodeGen->PrintStats();
    consumers.pop_back();
    consumers.insert(consumers.begin(), lastConsumer);
  }
  
  virtual void HandleTranslationUnit(clang::ASTContext &Context) {
    Visitor->TraverseDecl(Context.getTranslationUnitDecl());
  }
private:
  rewriter_exVisitor *Visitor;
  CompilerInstance *CI;
};

class rewriter_exAction : public PluginASTAction {
protected:
  ASTConsumer *CreateASTConsumer(CompilerInstance &CI, llvm::StringRef) {
    return new rewriter_exConsumer(&CI);
  }
  
  bool ParseArgs(const CompilerInstance &CI, const std::vector<std::string>& args) {
    for (unsigned i = 0, e = args.size(); i != e; ++i) {
      llvm::errs() << "rewriter_ex arg = " << args[i] << "\n";
      // Example error handling.
      if (args[i] == "-an-error") {
        DiagnosticsEngine &D = CI.getDiagnostics();
        unsigned DiagID = D.getCustomDiagID(
          DiagnosticsEngine::Error, "invalid argument '" + args[i] + "'");
        D.Report(DiagID);
        return false;
      }
    }
    if (args.size() && args[0] == "help")
      PrintHelp(llvm::errs());

    return true;
  }
  
  void PrintHelp(llvm::raw_ostream& ros) {
    ros << "Help for rewriter_ex plugin goes here\n";
  }
};

}

static FrontendPluginRegistry::Add<rewriter_exAction>
X("run-plugin", "switch the original value");
